# -*- coding: utf-8 -*-
"""
-------------------------------------------------
   Author :       galen
   date：          2018/5/21
-------------------------------------------------
   Description:
-------------------------------------------------
"""
import math

x_pi = 3.14159265358979324 * 3000.0 / 180.0
pi = 3.1415926535897932384626  # π
a = 6378245.0  # 长半轴
ee = 0.00669342162296594323  # 扁率


def gcj02_to_bd09(lng, lat):
    """
    Mars coordinate system (GCJ-02) to Baidu coordinate system (BD-09)
    Google, GAD - > Baidu
    :param lng:Mars lng
    :param lat:Mars lat
    :return:
    """
    z = math.sqrt(lng * lng + lat * lat) + 0.00002 * math.sin(lat * x_pi)
    theta = math.atan2(lat, lng) + 0.000003 * math.cos(lng * x_pi)
    bd_lng = z * math.cos(theta) + 0.0065
    bd_lat = z * math.sin(theta) + 0.006
    return [bd_lng, bd_lat]


def bd09_to_gcj02(bd_lon, bd_lat):
    """
    Baidu coordinate system (BD-09) to  Mars coordinate system (GCJ-02)
    Baidu - > Google, Gao De
    :param bd_lat:Baidu lng
    :param bd_lon:Baidu lat
    :return:
    """
    x = bd_lon - 0.0065
    y = bd_lat - 0.006
    z = math.sqrt(x * x + y * y) - 0.00002 * math.sin(y * x_pi)
    theta = math.atan2(y, x) - 0.000003 * math.cos(x * x_pi)
    gg_lng = z * math.cos(theta)
    gg_lat = z * math.sin(theta)
    return [gg_lng, gg_lat]


def wgs84_to_gcj02(lng, lat):
    """
    WGS84 to GCJ02 (Mars coordinate system)
    :param lng:WGS84 lng
    :param lat:WGS84 lat
    :return:
    """
    # if out_of_china(lng, lat):
    #     return lng, lat
    d_lat = _trans_form_lat(lng - 105.0, lat - 35.0)
    d_lng = _trans_form_lng(lng - 105.0, lat - 35.0)
    rad_lat = lat / 180.0 * pi
    magic = math.sin(rad_lat)
    magic = 1 - ee * magic * magic
    sqrt_magic = math.sqrt(magic)
    d_lat = (d_lat * 180.0) / ((a * (1 - ee)) / (magic * sqrt_magic) * pi)
    d_lng = (d_lng * 180.0) / (a / sqrt_magic * math.cos(rad_lat) * pi)
    mg_lat = lat + d_lat
    mg_lng = lng + d_lng
    return [mg_lng, mg_lat]


def gcj02_to_wgs84(lng, lat):
    """
    GCJ02 (Mars coordinate system) WGS84
    :param lng:Mars lng
    :param lat:Mars lat
    :return:
    """
    if out_of_china(lng, lat):
        return lng, lat
    d_lat = _trans_form_lat(lng - 105.0, lat - 35.0)
    d_lng = _trans_form_lng(lng - 105.0, lat - 35.0)
    rad_lat = lat / 180.0 * pi
    magic = math.sin(rad_lat)
    magic = 1 - ee * magic * magic
    sqrt_magic = math.sqrt(magic)
    d_lat = (d_lat * 180.0) / ((a * (1 - ee)) / (magic * sqrt_magic) * pi)
    d_lng = (d_lng * 180.0) / (a / sqrt_magic * math.cos(rad_lat) * pi)
    mg_lat = lat + d_lat
    mg_lng = lng + d_lng
    return [lng * 2 - mg_lng, lat * 2 - mg_lat]


def bd09_to_wgs84(bd_lng, bd_lat):
    lon, lat = bd09_to_gcj02(bd_lng, bd_lat)
    return gcj02_to_wgs84(lon, lat)


def wgs84_to_bd09(lng, lat):
    lng, lat = wgs84_to_gcj02(lng, lat)
    return gcj02_to_bd09(lng, lat)


def _trans_form_lat(lng, lat):
    ret = -100.0 + 2.0 * lng + 3.0 * lat + 0.2 * lat * lat + \
          0.1 * lng * lat + 0.2 * math.sqrt(math.fabs(lng))
    ret += (20.0 * math.sin(6.0 * lng * pi) + 20.0 *
            math.sin(2.0 * lng * pi)) * 2.0 / 3.0
    ret += (20.0 * math.sin(lat * pi) + 40.0 *
            math.sin(lat / 3.0 * pi)) * 2.0 / 3.0
    ret += (160.0 * math.sin(lat / 12.0 * pi) + 320 *
            math.sin(lat * pi / 30.0)) * 2.0 / 3.0
    return ret


def _trans_form_lng(lng, lat):
    ret = 300.0 + lng + 2.0 * lat + 0.1 * lng * lng + \
          0.1 * lng * lat + 0.1 * math.sqrt(math.fabs(lng))
    ret += (20.0 * math.sin(6.0 * lng * pi) + 20.0 *
            math.sin(2.0 * lng * pi)) * 2.0 / 3.0
    ret += (20.0 * math.sin(lng * pi) + 40.0 *
            math.sin(lng / 3.0 * pi)) * 2.0 / 3.0
    ret += (150.0 * math.sin(lng / 12.0 * pi) + 300.0 *
            math.sin(lng / 30.0 * pi)) * 2.0 / 3.0
    return ret


def out_of_china(lng, lat):
    """
    Whether or not it is in China
    :param lng:
    :param lat:
    :return:
    """
    return not (lng > 73.66 < 135.05 and 3.86 < lat < 53.55)


def conversion(coordinate_sys, lng, lat):
    lng = float(lng)
    lat = float(lat)
    if coordinate_sys.lower() == 'gcj02':
        # [mg_lng, mg_lat]
        coo = wgs84_to_gcj02(lng, lat)
        return str(coo[0]), str(coo[1])
    if coordinate_sys.lower() == 'bd09':
        #  [bd_lng, bd_lat]
        coo = wgs84_to_bd09(lng, lat)
        return str(coo[0]), str(coo[1])
    return str(lng), str(lat)



class Boundary:
    def __init__(self):
        self.name = ''
        self.relation_id = ''
        self.boundary_outer = ''
        self.boundary_inner = ''

    @property
    def info(self):
        data = {
            'name': self.name,
            "relation_id": self.relation_id,
            'boundary': {
                'outer': self.boundary_outer,
                'inner': self.boundary_inner
            }
        }
        return data



"""
-------------------------------------------------
   Author :       galen
   date：          2018/5/22
-------------------------------------------------
   Description:
-------------------------------------------------
"""
from collections import defaultdict


class Way:
    def __init__(self, value, reverse=False):
        self.value = value
        self.reverse = reverse


class Region:
    """
    Region object
    """

    def __init__(self):
        self._name = ''
        self._node_dict = defaultdict(dict)
        self._way_dict = defaultdict(list)
        self._area_outer = []
        self._area_inner = []
        self._order_outer = []
        self._order_inner = []

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, values):
        self._name = values

    @property
    def nodes(self):
        return self._node_dict

    @nodes.setter
    def nodes(self, values):
        self._node_dict = values

    @property
    def ways(self):
        return self._way_dict

    @ways.setter
    def ways(self, values):
        self._way_dict = values

    @property
    def area_outer(self):
        return self._area_outer

    @area_outer.setter
    def area_outer(self, values):
        self._area_outer = values

    @property
    def area_inner(self):
        return self._area_inner

    @area_inner.setter
    def area_inner(self, values):
        self._area_inner = values

    @property
    def order_outer(self):
        return self._order_outer

    @order_outer.setter
    def order_outer(self, values):
        self._order_outer = values

    @property
    def order_inner(self):
        return self._order_inner

    @order_inner.setter
    def order_inner(self, values):
        self._order_inner = values



"""
-------------------------------------------------
   Author :       galen
   date：          2018/5/21
-------------------------------------------------
   Description:
-------------------------------------------------
"""
from .transformation import conversion


class DataOutput:
    def __init__(self):
        self.csys = None
        self.coo_order = None
        self.region = None
        self.boundary = None

    def format_node(self, node):
        # {'lon': str(lon), 'lat': str(lat)}
        lng = self.region.nodes[node]['lon']
        lat = self.region.nodes[node]['lat']
        coordinate = lng + "," + lat
        if self.csys != 'wgs84':
            lng, lat = conversion(self.csys, lng, lat)
        if self.coo_order:
            coordinate = lat + "," + lng
        return coordinate

    def reshape_area(self, areas):
        area_str = ''
        for area in areas:
            ways_str = ''
            for way in area:
                way_values = self.region.ways[way.value]
                if way.reverse:
                    way_values.reverse()
                way_str = ';'.join(list(map(self.format_node, way_values))) + ';'
                ways_str += way_str
            area_str += ways_str + '='
        return area_str.strip('=')

    def reshape_data(self, region, boundary, csys=False, coo_order=False):
        self.csys = csys
        self.coo_order = coo_order
        self.region = region
        self.boundary = boundary
        self.boundary.boundary_outer = self.reshape_area(self.region.order_outer)
        self.boundary.boundary_inner = self.reshape_area(self.region.order_inner)



"""
-------------------------------------------------
   Author :       galen
   date：          2018/5/21
-------------------------------------------------
   Description:
                Parser nodes, ways, areas
-------------------------------------------------
"""

from .errors import BoundaryError
from .region import Way


class Parser:
    def __init__(self):
        self.region = None

    def combine(self, tem_boundary, ways, area, area_start, area_end):
        """
        tem_boundary, next_area, area_start, area_end
        """
        if ways[0] == area:
            next_area = ways[-1]
        else:
            next_area = ways[0]
        _area = self.region.ways[area]
        _next_area = self.region.ways[next_area]
        if area_end == _next_area[0]:
            tem_boundary.append(Way(next_area))
            return tem_boundary, next_area, area_start, _next_area[-1], _next_area[-1]
        elif area_end == _next_area[-1]:
            tem_boundary.append(Way(next_area, True))
            return tem_boundary, next_area, area_start, _next_area[0], _next_area[0]
        elif area_start == _area[-1]:
            tem_boundary.insert(0, Way(next_area))
            return tem_boundary, next_area, _next_area[0], area_end, _next_area[0]
        elif area_start == _area[-1]:
            tem_boundary.insert(0, Way(next_area, True))
            return tem_boundary, next_area, _next_area[-1], area_end, _next_area[-1]
        else:
            raise BoundaryError('Combine error')

    @staticmethod
    def count_se(se_node, area, ele):
        if ele in se_node:
            se_node[ele]['count'] += 1
            se_node[ele]['ways'].append(area)
        else:
            se_node[ele] = {'count': 1, "ways": [area]}
        return se_node

    def order_areas(self, areas):
        if len(areas) < 0:
            return
        se_node = {}
        for area in areas:
            way = self.region.ways[area]
            se_node = self.count_se(se_node, area, way[0])
            se_node = self.count_se(se_node, area, way[-1])
        for k, v in se_node.items():
            if v['count'] != 2:
                raise BoundaryError("Path can't closed")
        data = ('', '', '', '', [], [])
        next_area, start, end, next_node, boundaries, tem_boundary = data
        areas_len = len(areas)
        while areas_len > 0:
            if len(tem_boundary) == 0:
                next_area = areas[0]
                tem_boundary.append(Way(next_area))
                start = self.region.ways[next_area][0]
                end = self.region.ways[next_area][-1]
                next_node = end
            areas.remove(next_area)
            areas_len = len(areas)
            if start == end:
                boundaries.append(tem_boundary)
                tem_boundary = []
                continue
            ways = se_node[next_node]['ways']
            tem_boundary, next_area, start, end, next_node = self.combine(tem_boundary, ways, next_area, start, end)
            if start == end:
                boundaries.append(tem_boundary)
                tem_boundary = []
        return boundaries

    def format_coordinate(self, region):
        self.region = region
        self.region.order_outer = self.order_areas(self.region.area_outer)
        self.region.order_inner = self.order_areas(self.region.area_inner)

        # def get_data(self):
        #     return self.region



"""
-------------------------------------------------
   Author :       galen
   date：          2018/5/21
-------------------------------------------------
   Description:
   OpenStreetMap ：WGS84 coordinate system
   default coordinate order : lng,lat  coo_order = False
   boundary.info : all data
            name : boundary name
            relation_id :relation id
            boundary : including outer and inner
                    outer
                    inner
-------------------------------------------------
"""

from .boundary import Boundary
from .parsers import Parser
from .extractors import Extractor, NameExtractor
from .downloader import HtmlFetcher
from .errors import InvalidRelationId, PagesNotExist
from .outputformatters import DataOutput


class Crawler:
    def __init__(self):
        self.html_fetcher = HtmlFetcher()
        self.boundary = Boundary()
        self.parsers = Parser()
        self.outputer = DataOutput()

    def id_parse(self, relation_id, csys='wgs84', coo_order=False):
        """
        get Boundary,Conversion coordinate
        csys -> Coordinate System
        :param relation_id:
        :param csys: wgs84:wgs84; gcj02: Mars; bd09:Baidu
        :param coo_order: lng,lat  coo_order = False; lat,lng  coo_order = True;
        :return:
        """
        if not relation_id.isdigit():
            raise InvalidRelationId('Invalid id <{0}>'.format(str(relation_id)))
        url = 'https://www.openstreetmap.org/api/0.6/relation/{0}/full'.format(str(relation_id))
        html = self.html_fetcher.get_html(url)
        if len(html) == 0:
            raise PagesNotExist('Pages Not Found')
        extractor = Extractor(html, relation_id)
        region = extractor.get_region()
        self.boundary.name = region.name
        self.boundary.relation_id = relation_id
        self.parsers.format_coordinate(region)
        self.outputer.reshape_data(region, self.boundary, csys, coo_order)
        return self.boundary

    def name_parse(self, name, level, csys='wgs84', coo_order=False):
        """
        get boundary by name
        :param csys:
        :param coo_order:
        :param name: place name
        :param level: Administrative district level ,country state city county town
        :return:
        """
        url = 'https://www.openstreetmap.org/geocoder/search_osm_nominatim?query={0}'.format(str(name))
        html = self.html_fetcher.get_html(url, is_text=True)
        if len(html) == 0:
            raise PagesNotExist('Pages Not Found')
        extractor = NameExtractor(html, name, level)
        data_id, _ = extractor.extractor_id()
        if data_id is None:
            return self.boundary
        return self.id_parse(data_id, csys, coo_order)



"""
-------------------------------------------------
   Author :       galen
   date：          2018/5/21
-------------------------------------------------
   Description:
-------------------------------------------------
"""
from lxml import etree
from collections import defaultdict
from .region import Region


class Extractor:
    def __init__(self, html, relation_id):
        self.selector = etree.HTML(html)
        self.relation_id = relation_id
        self.region = Region()

    def get_nodes(self):
        node_dict = defaultdict(dict)
        node_list = self.selector.xpath('//node')
        for node in node_list:
            node_id = node.xpath('@id')[0]
            lon = node.xpath('@lon')[0]
            lat = node.xpath('@lat')[0]
            node_dict[str(node_id)] = {'lon': str(lon), 'lat': str(lat)}
        self.region.nodes = node_dict

    def get_ways(self):
        way_dict = defaultdict(list)
        way_list = self.selector.xpath('//way')
        for way in way_list:
            way_id = way.xpath('@id')[0]
            way_node_ele = way.xpath('nd/@ref')
            way_node = list(map(lambda x: str(x), way_node_ele))
            way_dict[str(way_id)] = way_node
        self.region.ways = way_dict

    def get_name(self):
        name = self.selector.xpath('//relation[@id="{0}"]/tag[@k="name:en"]/@v'.format(self.relation_id))
        if len(name) == 0:
            name = self.selector.xpath('//relation[@id="{0}"]/tag[@k="name"]/@v'.format(self.relation_id))
        self.region.name = str(name[0])

    def get_area(self):
        # role = inner outer
        _outer = self.selector.xpath(
            '//relation[@id="{0}"]/member[@role="outer"]/@ref'.format(self.relation_id))
        _inner = self.selector.xpath(
            '//relation[@id="{0}"]/member[@role="inner"]/@ref'.format(self.relation_id))
        self.region.area_outer = list(map(lambda x: str(x), _outer))
        self.region.area_inner = list(map(lambda x: str(x), _inner))

    def get_region(self):
        self.get_name()
        self.get_nodes()
        self.get_ways()
        self.get_area()
        return self.region


class NameExtractor:
    def __init__(self, html, name, level):
        self.selector = etree.HTML(html)
        self.name = name
        self.level = level

    def extractor_id(self):
        search_list = self.selector.xpath('//ul[@class="results-list"]/li')
        dist_dict = {
            'country': 'Country',
            'state': 'State',
            'city': 'Region Boundary',
            'county': 'County Boundary',
        }
        for search in search_list:
            data_type = search.xpath('p/a/@data-type')[0]
            if data_type != 'relation':
                continue
            prefix = search.xpath('p/a/@data-prefix')[0]
            if dist_dict[self.level] != prefix:
                continue
            name = search.xpath('p/a/@data-name')[0]
            data_id = search.xpath('p/a/@data-id')[0]
            return data_id, name
        return None, None



"""
-------------------------------------------------
   Author :       galen
   date：          2018/5/21
-------------------------------------------------
   Description:
-------------------------------------------------
"""


class OsmBaseException(Exception):
    def __init__(self, message):
        super(OsmBaseException, self).__init__()
        self._message = message

    def __str__(self):
        return '{0}.'.format(self._message)

    def __repr__(self):
        return self.__str__()

    def __unicode__(self):
        return '{0}.'.format(self._message)

    @property
    def msg(self):
        return self._message


class PagesNotExist(OsmBaseException):
    pass


class InvalidRelationId(OsmBaseException):
    pass


class BoundaryError(OsmBaseException):
    pass



# -*- coding: utf-8 -*-


from setuptools import setup, find_packages
from openstreemap import __version__

VERSION = __version__
# with open("README.md", "r") as fh:
#     long_description = fh.read()
long_description = 'Api For OpenStreetMap coordinates'
setup(
    name='openstreetmap',
    version=VERSION,
    description='OpenStreetMap coordinates',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache-2.0",
    author='galen',
    author_email='mywayking@icloud.com',
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    url='https://github.com/Mywayking/openstreetmap.git',
    keywords='openstreetmap',
    packages=find_packages(),
    install_requires=[
        'lxml',
        'requests',
    ],
    python_requires='>=3.4',
)



"""
-------------------------------------------------
   Author :       galen
   date：          2018/5/21
-------------------------------------------------
   Description: downloader
-------------------------------------------------
"""
import random
import requests

HTTP_DEFAULT_TIMEOUT = 20
BROWSER_USER_AGENT = [
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1985.67 Safari/537.36",
    "Mozilla/5.0 (X11; OpenBSD i386) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1985.125 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1944.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.3319.102 Safari/537.36",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.2309.372 Safari/537.36",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.2117.157 Safari/537.36",
    "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.1916.47 Safari/537.36",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1866.237 Safari/537.36",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1847.137 Safari/4E423F",
    "Mozilla/5.0 (Windows NT 5.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1847.116 Safari/537.36 Mozilla/5.0 (iPad; U; CPU OS 3_2 like Mac OS X; en-us) AppleWebKit/531.21.10 (KHTML, like Gecko) Version/4.0.4 Mobile/7B334b Safari/531.21.10",
    "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/33.0.1750.517 Safari/537.36",
    "Mozilla/5.0 (Windows NT 6.2; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/32.0.1667.0 Safari/537.36",
]


class HtmlFetcher:
    def __init__(self):
        self.http_timeout = HTTP_DEFAULT_TIMEOUT
        self.headers = {'User-agent': random.choice(BROWSER_USER_AGENT)}

    def get_html(self, url, is_text=False):
        # do request:
        req = requests.get(url, headers=self.headers, timeout=self.http_timeout)
        req.raise_for_status()
        if is_text:
            return req.text
        return req.content
